/////////////////////////////////////////////////////////////////////////////
//
// Windows MFC Glk Libraries
//
// GlkDLL
// Glk dll entry and exit points
//
/////////////////////////////////////////////////////////////////////////////

#ifndef WINGLK_DLL_H_
#define WINGLK_DLL_H_

extern "C"
{
#include "glk.h"
#include "gi_blorb.h"
#include "gi_dispa.h"

extern void (*InterruptFn)(void);
extern gidispatch_rock_t (*RegisterObjFn)(void *obj, glui32 objclass);
extern void (*UnregisterObjFn)(void *obj, glui32 objclass, gidispatch_rock_t objrock);
extern gidispatch_rock_t (*RegisterArrFn)(void *array, glui32 len, char *typecode);
extern void (*UnregisterArrFn)(void *array, glui32 len, char *typecode, gidispatch_rock_t objrock);
}

#include "GlkGraphic.h"
#include "GlkSound.h"
#include "Resource.h"

/////////////////////////////////////////////////////////////////////////////
// CGlkApp
/////////////////////////////////////////////////////////////////////////////

class CGlkApp : public CWinApp
{
public:
	CGlkApp();

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CGlkApp)
	public:
	virtual BOOL InitInstance();
	virtual int ExitInstance();
	//}}AFX_VIRTUAL

	//{{AFX_MSG(CGlkApp)
		// NOTE - the ClassWizard will add and remove member functions here.
		//    DO NOT EDIT what you see in these blocks of generated code !
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()

public:
	LOGFONT* GetPropFont(void) { return &m_PropFont; };
	LOGFONT* GetFixedFont(void) { return &m_FixedFont; };
	CRect& GetWindowRect(void) { return m_WindowRect; };
	int& GetWindowState(void) { return m_iWindowState; };

	void ReadSettings(void);
	void WriteSettings(void);
	void LoadConfigFile(const char* pszConfigName);
	void SetSaveOptions(bool bSave) { m_bSaveSettings = bSave; };

	bool GetWindowBorders(void) { return m_bWindowBorders; };
	bool SetWindowBorders(bool bBorders);

	bool GetEnableGUI(void) { return m_bEnableGUI; };
	bool SetEnableGUI(bool bEnableGUI);

	bool GetStyleHints(void) { return m_bStyleHints; };
	void SetStyleHints(bool bHints) { m_bStyleHints = bHints; };

	bool GetCanSpeak(void) { return m_bSpeak; };
	void SetCanSpeak(bool bSpeak) { m_bSpeak = bSpeak; };

	GUID GetVoice(void) { return m_VoiceGUID; };
	void SetVoice(GUID VoiceGUID) { m_VoiceGUID = VoiceGUID; };

	bool GetColourLinks(void) { return m_bColourLinks; };
	void SetColourLinks(bool bColour) { m_bColourLinks = bColour; };
	bool GetUnderlineLinks(void) { return m_bUnderlineLinks; };
	void SetUnderlineLinks(bool bUnderline) { m_bUnderlineLinks = bUnderline; };
	COLORREF GetCustomLinkColour(void) { return m_CustomLinkColour; };
	void SetCustomLinkColour(COLORREF Colour) { m_CustomLinkColour = Colour; };
	COLORREF GetLinkColour(void);

	CString& GetAppNameRef(void) { return m_strAppName; };
	CString& GetAppTitleRef(void) { return m_strAppTitle; };
	CString& GetResourceDirRef(void) { return m_strResDir; };
	CString& GetInitialDirRef(void) { return m_strInitialDir; };

	UINT GetUserGuiID(void) { return m_UserGuiID; };
	void SetUserGuiID(UINT id) { m_UserGuiID = id; };

	giblorb_map_t*& GetBlorbMapRef(void) { return m_pBlorbMap; };
	strid_t& GetBlorbFileRef(void) { return m_BlorbFile; };

	bool CreateMainWindow(void);
	bool EventQueuesEmpty(void);
	void GetNextEvent(event_t* pEvent, bool bPoll);
	void AddEvent(glui32 Type, winid_t Win, glui32 Value1, glui32 Value2);
	void MessagePump(BOOL bWait);

	CWinGlkGraphic* LoadGraphic(int iNumber, BOOL bLoad);
	CWinGlkSound* LoadSound(int iNumber);

	bool CanSpeakChar(char c);
	void Speak(LPCTSTR pszText);
	void InitSoundEngine(void);

protected:
	CString FileName(LPCTSTR pszPrefix,int iIndex,LPCTSTR pszSuffix);
	void DeleteOldTempFiles(void);
	void LoadInternationalResources(void);

protected:
	bool m_bSettingsRead;
	bool m_bSaveSettings;

	LOGFONT m_PropFont;
	LOGFONT m_FixedFont;
	CRect m_WindowRect;
	int m_iWindowState;
	bool m_bWindowBorders;
	bool m_bWindowFrame;
	bool m_bEnableGUI;
	bool m_bStyleHints;

	bool m_bSpeak;
	GUID m_VoiceGUID;

	bool m_bColourLinks;
	bool m_bUnderlineLinks;
	COLORREF m_CustomLinkColour;

	CString m_strAppName;
	CString m_strAppTitle;
	CString m_strResDir;
	CString m_strInitialDir;

	UINT m_UserGuiID;

	CArray<event_t,event_t&> InputEvents;
	CArray<event_t,event_t&> TimerEvents;
	CArray<event_t,event_t&> ArrangeEvents;
	CArray<event_t,event_t&> RedrawEvents;
	CArray<event_t,event_t&> SoundEvents;

	strid_t m_BlorbFile;
	giblorb_map_t* m_pBlorbMap;
};

#endif // WINGLK_DLL_H_
