/////////////////////////////////////////////////////////////////////////////
//
// Windows MFC Glk Libraries
//
// GlkMainWnd
// Main window of the Glk application
//
/////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "GlkDll.h"
#include "GlkMainWnd.h"
#include "GlkDialogs.h"
#include "GlkSound.h"
#include "GlkTalk.h"
#include "GlkWindowGfx.h"
#include "GlkWindowTextBuffer.h"
#include "WinGlk.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CWinGlkViewWnd window, derived from the base CWnd class
/////////////////////////////////////////////////////////////////////////////

BEGIN_MESSAGE_MAP(CWinGlkViewWnd, CWnd)
	//{{AFX_MSG_MAP(CWinGlkViewWnd)
	ON_WM_SIZE()
	ON_WM_PAINT()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void CWinGlkViewWnd::OnSize(UINT nType, int cx, int cy) 
{
	CWnd::OnSize(nType,cx,cy);
	SizeWindows();
	Invalidate();
}

void CWinGlkViewWnd::OnPaint() 
{
	// Must always be present here so that ::BeginPaint() and
	// ::EndPaint() get called in the constructor and destructor
	CPaintDC dc(this);

	if (CWinGlkWnd::GetMainWindow() == NULL)
	{
		CRect Client;
		GetClientRect(Client);
		dc.FillSolidRect(Client,::GetSysColor(COLOR_WINDOW));
	}
}

void CWinGlkViewWnd::SizeWindows(void)
{
	if (CWinGlkWnd::GetMainWindow())
	{
		CRect ClientRect;
		GetClientRect(ClientRect);

		CWinGlkWnd::GetMainWindow()->SizeWindow(&ClientRect);
		((CGlkApp*)AfxGetApp())->AddEvent(evtype_Arrange,0,0,0);
	}	
}

/////////////////////////////////////////////////////////////////////////////
// CWinGlkMainWnd frame window, derived from the CFrameWnd class
/////////////////////////////////////////////////////////////////////////////

static UINT Indicators[] =
{
	ID_SEPARATOR,
	ID_INDICATOR_CAPS,
	ID_INDICATOR_NUM,
	ID_INDICATOR_SCRL,
};

CWinGlkMainWnd::CWinGlkMainWnd()
{
}

CWinGlkMainWnd::~CWinGlkMainWnd()
{
}

BEGIN_MESSAGE_MAP(CWinGlkMainWnd, CFrameWnd)
	//{{AFX_MSG_MAP(CWinGlkMainWnd)
	ON_WM_CHAR()
	ON_WM_KEYDOWN()
	ON_WM_SYSCOMMAND()
	ON_WM_DESTROY()
	ON_WM_SETTINGCHANGE()
	ON_WM_INITMENUPOPUP()
	ON_COMMAND(IDM_SYS_SCROLLBACK, OnScrollback)
	ON_UPDATE_COMMAND_UI(IDM_SYS_SCROLLBACK, OnUpdateScrollback)
	ON_COMMAND(IDM_FONT_PROPORTIONAL, OnFontProportional)
	ON_COMMAND(IDM_FONT_FIXED, OnFontFixed)
	ON_COMMAND(IDM_SYS_OPTIONS, OnOptions)
	ON_COMMAND(IDM_SYS_ABOUT, OnAbout)
	ON_WM_TIMER()
	//}}AFX_MSG_MAP
	ON_WM_MEASUREITEM()
	ON_WM_MENUCHAR()
	ON_WM_SYSKEYDOWN()
	ON_NOTIFY_EX_RANGE(TTN_NEEDTEXTW,0,0xFFFF,OnToolTipText)
	ON_NOTIFY_EX_RANGE(TTN_NEEDTEXTA,0,0xFFFF,OnToolTipText)
END_MESSAGE_MAP()

bool CWinGlkMainWnd::Create(bool bFrame)
{
	CGlkApp* pApp = (CGlkApp*)AfxGetApp();
	CRect rPlace(pApp->GetWindowRect());
	if (rPlace.Width() == 0)
	{
		int iWidth = GetSystemMetrics(SM_CXFULLSCREEN);
		int iHeight = GetSystemMetrics(SM_CYFULLSCREEN);
		rPlace.right = (iWidth*6)/7;
		rPlace.bottom = (iHeight*6)/7;
	}

	// Create the frame
	DWORD dwStyle = bFrame ? WS_OVERLAPPEDWINDOW : WS_POPUP;
	if (CreateEx(0,AfxRegisterWndClass(0),pApp->GetAppTitleRef(),
		dwStyle,rPlace,NULL,0,NULL) == FALSE)
	{
		return FALSE;
	}

	// Set the mask for the frame, if any
	SetWindowRgn(m_Mask,FALSE);

	CRect InitRect;
	GetClientRect(InitRect);

	DWORD dwExStyle = WS_EX_CLIENTEDGE;
	if (pApp->GetWindowBorders() || (bFrame == false))
		dwExStyle = 0;

	// Create the view, which contains all the Glk windows
	if (m_View.CreateEx(dwExStyle,
		AfxRegisterWndClass(0,pApp->LoadStandardCursor(IDC_ARROW)),
		NULL,WS_CHILD|WS_VISIBLE,InitRect,this,AFX_IDW_PANE_FIRST,NULL) == FALSE)
	{
		return FALSE;
	}

	// Set up the system menu
	CString strMenuItem;
	m_FontMenu.LoadMenu(IDR_FONTS);
	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu)
	{
		strMenuItem.LoadString(IDS_MENU_SCROLLBACK);
		pSysMenu->InsertMenu(6,MF_BYPOSITION|MF_STRING,IDM_SYS_SCROLLBACK,strMenuItem);

		strMenuItem.LoadString(IDS_MENU_FONTS);
		pSysMenu->InsertMenu(7,MF_BYPOSITION|MF_STRING|MF_POPUP,(UINT)m_FontMenu.m_hMenu,strMenuItem);

		strMenuItem.LoadString(IDS_MENU_OPTIONS);
		pSysMenu->InsertMenu(8,MF_BYPOSITION|MF_STRING,IDM_SYS_OPTIONS,strMenuItem);

		strMenuItem.LoadString(IDS_MENU_ABOUT);
		pSysMenu->InsertMenu(9,MF_BYPOSITION|MF_STRING,IDM_SYS_ABOUT,strMenuItem);
		pSysMenu->InsertMenu(10,MF_BYPOSITION|MF_SEPARATOR);
	}

	// Set up the main menu
	UINT MenuID = pApp->GetUserGuiID();
	if (MenuID != 0)
		m_Menus.LoadMenu(MenuID);

	// If no menu has been loaded, create a blank menu bar
	if (m_Menus.GetSafeHmenu() == 0)
		m_Menus.CreateMenu();

	const char* pszGlkMenu = "&Glk";
	BCMenu* pGlkMenu = m_Menus.AppendODPopupMenu(pszGlkMenu);
	m_Menus.ModifyMenu(m_Menus.GetMenuItemCount()-1,MF_POPUP|MF_BYPOSITION,
		(UINT)pGlkMenu->GetSafeHmenu(),pszGlkMenu);

	strMenuItem.LoadString(IDS_MENU_SCROLLBACK);
	pGlkMenu->AppendMenu(MF_STRING,IDM_SYS_SCROLLBACK,strMenuItem);

	strMenuItem.LoadString(IDS_MENU_FONTS);
	BCMenu* pFontMenu = pGlkMenu->AppendODPopupMenu(strMenuItem);
	strMenuItem.LoadString(IDS_MENU_PROPFONT);
	pFontMenu->AppendMenu(MF_STRING,IDM_FONT_PROPORTIONAL,strMenuItem);
	strMenuItem.LoadString(IDS_MENU_FIXEDFONT);
	pFontMenu->AppendMenu(MF_STRING,IDM_FONT_FIXED,strMenuItem);

	strMenuItem.LoadString(IDS_MENU_OPTIONS);
	pGlkMenu->AppendMenu(MF_STRING,IDM_SYS_OPTIONS,strMenuItem);

	strMenuItem.LoadString(IDS_MENU_ABOUT);
	pGlkMenu->AppendMenu(MF_STRING,IDM_SYS_ABOUT,strMenuItem);

	strMenuItem.LoadString(IDS_MENU_EXIT);
	pGlkMenu->AppendMenu(MF_SEPARATOR,0,NULL);
	pGlkMenu->AppendMenu(MF_STRING,ID_APP_EXIT,strMenuItem);

	// Load the images for the menus
	if (MenuID != 0)
		m_Menus.LoadToolbar(MenuID);
	m_Menus.LoadToolbar(IDR_GLK);

	// Set up the status bar
	if ((m_StatusBar.Create(this,WS_CHILD|CBRS_BOTTOM) == FALSE) ||
		(m_StatusBar.SetIndicators(Indicators,sizeof(Indicators)/sizeof(UINT)) == FALSE))
	{
		return FALSE;
	}

	// Load the user toolbar
	UINT ToolbarID = pApp->GetUserGuiID();
	if (ToolbarID != 0)
	{
		if (m_UserToolBar.Create(this))
		{
			if (m_UserToolBar.LoadToolBar(ToolbarID))
			{
				m_UserToolBar.SetBarStyle(m_UserToolBar.GetBarStyle() |
					CBRS_TOOLTIPS|CBRS_FLYBY|CBRS_SIZE_DYNAMIC);
				m_UserToolBar.EnableDocking(CBRS_ALIGN_ANY);
			}
			else
				m_UserToolBar.DestroyWindow();
		}
	}

	// Load the standard Glk toolbar
	if (m_GlkToolBar.Create(this) == FALSE)
		return FALSE;
	if (m_GlkToolBar.LoadToolBar(IDR_GLK) == FALSE)
		return FALSE;
	m_GlkToolBar.SetBarStyle(m_GlkToolBar.GetBarStyle() |
		CBRS_TOOLTIPS|CBRS_FLYBY|CBRS_SIZE_DYNAMIC);
	m_GlkToolBar.EnableDocking(CBRS_ALIGN_ANY);

	// Dock the toolbars
	EnableDocking(CBRS_ALIGN_ANY);
	if (m_UserToolBar.GetSafeHwnd())
		DockControlBar(&m_UserToolBar,false);
	DockControlBar(&m_GlkToolBar,true);

	// Load the icon
	HICON Icon = NULL;
	if (pApp->GetUserGuiID() != 0)
	{
		Icon = ::LoadIcon(::GetModuleHandle(NULL),
			MAKEINTRESOURCE(pApp->GetUserGuiID()));
	}
	if (Icon == NULL)
		Icon = pApp->LoadIcon(IDR_GLK);
	if (Icon == NULL)
		return FALSE;
	SetIcon(Icon,TRUE);

	// Load the accelerators
	LoadAccelTable(MAKEINTRESOURCE(IDR_GLK));

	// Enable or disable GUI elements
	SetGUI(pApp->GetEnableGUI());

	// Center the frame window
	if ((pApp->GetWindowRect().Width() == 0) || (bFrame == false))
		CenterWindow();

	// Start a regular 1/2 second pulse timer
	SetTimer(PulseTimer,500,NULL);
	return TRUE;
}

// MFC doesn't directly provide a way to dock two control bars on the
// same row, so this function is called instead of the usual MFC
// CFrameWnd::DockControlBar().
void CWinGlkMainWnd::DockControlBar(CControlBar* pBar, bool bLast)
{
	if ((pBar->m_dwStyle & CBRS_ALIGN_ANY) == CBRS_ALIGN_TOP)
	{
		CDockBar* pDockBar = (CDockBar*)GetControlBar(AFX_IDW_DOCKBAR_TOP);

		pDockBar->m_dwStyle &= ~(CBRS_SIZE_FIXED | CBRS_SIZE_DYNAMIC);
		pDockBar->m_dwStyle |= pBar->m_dwStyle & (CBRS_SIZE_FIXED | CBRS_SIZE_DYNAMIC);

		DWORD dwStyle = pBar->GetBarStyle();
		dwStyle &= ~(CBRS_ALIGN_ANY);
		dwStyle |= (pDockBar->m_dwStyle & CBRS_ALIGN_ANY) | CBRS_BORDER_ANY;
		if (pDockBar->m_bFloating)
			dwStyle |= CBRS_FLOATING;
		else
			dwStyle &= ~CBRS_FLOATING;
		pBar->SetBarStyle(dwStyle);

		pDockBar->m_arrBars.Add(pBar);
		if (bLast)
			pDockBar->m_arrBars.Add(NULL);

		pBar->SetWindowPos(NULL,-2,-2,0,0,
			SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOCOPYBITS);

		if (pBar->GetParent() != pDockBar)
			pBar->SetParent(pDockBar);
		pBar->m_pDockBar = pDockBar;

		pDockBar->RemovePlaceHolder(pBar);
		DelayRecalcLayout();
	}
	else
		CFrameWnd::DockControlBar(pBar);
}

BOOL CWinGlkMainWnd::PreCreateWindow(CREATESTRUCT& cs) 
{
	BOOL bPreCreate = CFrameWnd::PreCreateWindow(cs);
	cs.dwExStyle &= ~WS_EX_CLIENTEDGE;
	return bPreCreate;
}

void CWinGlkMainWnd::OnDestroy() 
{
	KillTimer(PulseTimer);
	KillTimer(GlkTimer);

	WINDOWPLACEMENT Place;
	GetWindowPlacement(&Place);

	CGlkApp* pApp = (CGlkApp*)AfxGetApp();
	pApp->GetWindowRect() = Place.rcNormalPosition;
	pApp->GetWindowState() = Place.showCmd;

	CFrameWnd::OnDestroy();
}

BOOL CWinGlkMainWnd::OnCmdMsg(UINT nID, int nCode, void* pExtra, AFX_CMDHANDLERINFO* pHandlerInfo) 
{
	CMap<CWinGlkWnd*,CWinGlkWnd*,int,int>& WindowMap = CWinGlkWnd::GetWindowMap();
	CWinGlkWnd* pWnd = NULL;
	int i;

	POSITION MapPos = WindowMap.GetStartPosition();
	while (MapPos)
	{
		WindowMap.GetNextAssoc(MapPos,pWnd,i);
		if (pWnd->OnCmdMsg(nID,nCode,pExtra,pHandlerInfo))
			return TRUE;
	}

	BOOL bCmdProcessed = CFrameWnd::OnCmdMsg(nID,nCode,pExtra,pHandlerInfo);
	if (bCmdProcessed == FALSE)
	{
		// Check for commands from the user menu and toolbar
		switch (nCode)
		{
		case CN_COMMAND:
			// Ignore standard MFC identifiers
			if (nID < 0xE000)
			{
				// Add an input event
				((CGlkApp*)AfxGetApp())->AddEvent(winglk_evtype_GuiInput,0,nID,0);
				bCmdProcessed = TRUE;
			}
			break;
		case CN_UPDATE_COMMAND_UI:
			{
				CCmdUI* pCmdUI = (CCmdUI*)pExtra;
				if (pCmdUI != NULL)
				{
					if (pCmdUI->m_pOther != NULL)
					{
						if (pCmdUI->m_pOther->IsKindOf(RUNTIME_CLASS(CToolBar)))
						{
							// User toolbar buttons are always enabled
							pCmdUI->Enable();
							bCmdProcessed = TRUE;
						}
					}
					else if (pCmdUI->m_pMenu != NULL)
					{
						// User menu items are always enabled
						pCmdUI->Enable();
						bCmdProcessed = TRUE;
					}
				}
			}
			break;
		}
	}
	return bCmdProcessed;
}

void CWinGlkMainWnd::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if (CWinGlkWnd::GetExiting())
	{
		::PostQuitMessage(0);
		return;
	}
	CFrameWnd::OnChar(nChar,nRepCnt,nFlags);

	CWinGlkWnd* pActiveWnd = CWinGlkWnd::GetActiveWindow();
	if (pActiveWnd)
	{
		switch (nChar)
		{
		case '\b':
		case '\r':
			pActiveWnd->InputChar(nChar);
			break;
		default:
			if ((nChar >= 32 && nChar <= 126) || (nChar >= 160 && nChar <= 255))
				pActiveWnd->InputChar(nChar);
			break;
		}
	}
}

void CWinGlkMainWnd::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if (CWinGlkWnd::GetExiting())
	{
		::PostQuitMessage(0);
		return;
	}
	CFrameWnd::OnKeyDown(nChar,nRepCnt,nFlags);

	CWinGlkWnd* pActiveWnd = CWinGlkWnd::GetActiveWindow();
	if (pActiveWnd)
	{
		switch (nChar)
		{
		case VK_TAB:
			{
				CWnd* pNextWnd = pActiveWnd;
				bool bNext = false;
				do
				{
					if (::GetKeyState(VK_SHIFT) & 0x8000)
					{
						pNextWnd = pNextWnd->GetWindow(GW_HWNDPREV);
						if (pNextWnd == NULL)
							pNextWnd = pActiveWnd->GetWindow(GW_HWNDLAST);
					}
					else
					{
						pNextWnd = pNextWnd->GetWindow(GW_HWNDNEXT);
						if (pNextWnd == NULL)
							pNextWnd = pActiveWnd->GetWindow(GW_HWNDFIRST);
					}

					if (pNextWnd == pActiveWnd)
						bNext = true;

					if (pNextWnd)
					{
						if (pNextWnd->IsKindOf(RUNTIME_CLASS(CWinGlkWnd)))
						{
							if (((CWinGlkWnd*)pNextWnd)->InputPending() == true)
							bNext = true;
						}
					}
					else
						bNext = true;
				}
				while (bNext == false);

				if (pNextWnd)
					((CWinGlkWnd*)pNextWnd)->SetActiveWindow();
			}
			break;
		case VK_LEFT:		// Cursor left
		case VK_RIGHT:	// Cursor right
		case VK_UP:			// Cursor up
		case VK_DOWN:		// Cursor down
		case VK_HOME:		// Home
		case VK_END:		// End
		case VK_DELETE:	// Delete
		case VK_ESCAPE:	// Escape
		case VK_PRIOR:	// Page up
		case VK_NEXT:		// Page down
		case VK_F1:			// Function keys...
		case VK_F2:
		case VK_F3:
		case VK_F4:
		case VK_F5:
		case VK_F6:
		case VK_F7:
		case VK_F8:
		case VK_F9:
		case VK_F10:
		case VK_F11:
		case VK_F12:
			pActiveWnd->InputChar(nChar + 0x100);
			break;
		}
	}
}

// F10 gets passed to the application via WM_SYSKEYDOWN as it is
// the legacy window menu key (like Alt).
void CWinGlkMainWnd::OnSysKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if (CWinGlkWnd::GetExiting())
	{
		::PostQuitMessage(0);
		return;
	}

	if (nChar == VK_F10)
	{
		CWinGlkWnd* pActiveWnd = CWinGlkWnd::GetActiveWindow();
		if (pActiveWnd)
			pActiveWnd->InputChar(nChar + 0x100);
	}
	else
		CFrameWnd::OnSysKeyDown(nChar,nRepCnt,nFlags);
}

void CWinGlkMainWnd::OnSysCommand(UINT nID, LPARAM lParam) 
{
	switch (nID&0xFFF0)
	{
	case IDM_SYS_SCROLLBACK:
		OnScrollback();
		break;
	case IDM_SYS_ABOUT:
		OnAbout();
		break;
	case IDM_FONT_PROPORTIONAL:
		ChangeFont(false);
		break;
	case IDM_FONT_FIXED:
		ChangeFont(true);
		break;
	case IDM_SYS_OPTIONS:
		OnOptions();
		break;
	}
	CFrameWnd::OnSysCommand(nID, lParam);
}

void CWinGlkMainWnd::OnTimer(UINT nIDEvent) 
{
	CFrameWnd::OnTimer(nIDEvent);

	switch (nIDEvent)
	{
	case PulseTimer:
		{
			for (CWinGlkSndChannel* pSnd = CWinGlkSndChannel::IterateChannels(NULL,NULL); pSnd; pSnd = CWinGlkSndChannel::IterateChannels(pSnd,NULL))
				pSnd->TimerPulse();
		}
		break;
	case GlkTimer:
		if (CWinGlkWnd::GetMainWindow())
			((CGlkApp*)AfxGetApp())->AddEvent(evtype_Timer,0,0,0);
		break;
	}
}

void CWinGlkMainWnd::OnSettingChange(UINT uFlags, LPCTSTR lpszSection) 
{
	CFrameWnd::OnSettingChange(uFlags, lpszSection);

	CMap<CWinGlkWnd*,CWinGlkWnd*,int,int>& WindowMap = CWinGlkWnd::GetWindowMap();
	CWinGlkWnd* pWnd = NULL;
	int i;

	// Clear all graphics windows
	POSITION MapPos = WindowMap.GetStartPosition();
	while (MapPos)
	{
		WindowMap.GetNextAssoc(MapPos,pWnd,i);
		if (pWnd->IsKindOf(RUNTIME_CLASS(CWinGlkWndGraphics)))
			pWnd->ClearWindow();
	}

	if (CWinGlkWnd::GetMainWindow())
		((CGlkApp*)AfxGetApp())->AddEvent(evtype_Redraw,0,0,0);
}

void CWinGlkMainWnd::OnMeasureItem(int nIDCtl, LPMEASUREITEMSTRUCT lpMeasureItemStruct) 
{
	bool bSetFlag = false;
	if (lpMeasureItemStruct->CtlType == ODT_MENU)
	{
		if (::IsMenu((HMENU)lpMeasureItemStruct->itemID))
		{
			CMenu* pMenu = CMenu::FromHandle((HMENU)lpMeasureItemStruct->itemID);
			if (BCMenu::IsMenu(pMenu))
			{
				m_Menus.MeasureItem(lpMeasureItemStruct);
				bSetFlag = true;
			}
		}
	}
	if (bSetFlag == false)
		CFrameWnd::OnMeasureItem(nIDCtl,lpMeasureItemStruct);
}

LRESULT CWinGlkMainWnd::OnMenuChar(UINT nChar, UINT nFlags, CMenu* pMenu) 
{
	LRESULT lResult;

	if (BCMenu::IsMenu(pMenu))
		lResult = BCMenu::FindKeyboardShortcut(nChar,nFlags,pMenu);
	else
		lResult = CFrameWnd::OnMenuChar(nChar,nFlags,pMenu);
	return lResult;
}

void CWinGlkMainWnd::OnInitMenuPopup(CMenu* pPopupMenu, UINT nIndex, BOOL bSysMenu)
{
	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pPopupMenu->GetSafeHmenu() != pSysMenu->GetSafeHmenu())
	{
		CFrameWnd::OnInitMenuPopup(pPopupMenu,nIndex,bSysMenu);
		if (BCMenu::IsMenu(pPopupMenu))
			BCMenu::UpdateMenu(pPopupMenu);
	}
}

BOOL CWinGlkMainWnd::OnToolTipText(UINT, NMHDR* pNMHDR, LRESULT* pResult)
{
	TOOLTIPTEXTA* pTTTA = (TOOLTIPTEXTA*)pNMHDR;
	TOOLTIPTEXTW* pTTTW = (TOOLTIPTEXTW*)pNMHDR;
	TCHAR szFullText[256];
	CString strTipText;
	UINT nID = pNMHDR->idFrom;
	if (pNMHDR->code == TTN_NEEDTEXTA && (pTTTA->uFlags & TTF_IDISHWND) ||
			pNMHDR->code == TTN_NEEDTEXTW && (pTTTW->uFlags & TTF_IDISHWND))
	{
		nID = (UINT)(WORD)::GetDlgCtrlID((HWND)nID);
	}

	if (nID != 0)
	{
		if (AfxLoadString(nID,szFullText) == 0)
		{
			LPCTSTR lpszName = MAKEINTRESOURCE((nID>>4)+1);
			if (::FindResource(NULL,lpszName,RT_STRING) != NULL)
				::LoadString(NULL,nID,szFullText,255);
		}
		AfxExtractSubString(strTipText,szFullText,1,'\n');
	}

	if (pNMHDR->code == TTN_NEEDTEXTA)
	{
		lstrcpyn(pTTTA->szText,strTipText,
			sizeof(pTTTA->szText)/sizeof(pTTTA->szText[0]));
	}
	else
	{
		_mbstowcsz(pTTTW->szText,strTipText,
			sizeof(pTTTW->szText)/sizeof(pTTTW->szText[0]));
	}
	*pResult = 0;

	::SetWindowPos(pNMHDR->hwndFrom,HWND_TOP,0,0,0,0,
		SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE|SWP_NOOWNERZORDER);
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// Command handlers
/////////////////////////////////////////////////////////////////////////////

void CWinGlkMainWnd::OnScrollback() 
{
	CWinGlkWnd* pActiveWnd = CWinGlkWnd::GetActiveWindow();
	if (pActiveWnd)
		pActiveWnd->Scrollback();
}

void CWinGlkMainWnd::OnUpdateScrollback(CCmdUI* pCmdUI) 
{
	BOOL bEnable = FALSE;
	CWinGlkWnd* pWnd = CWinGlkWnd::GetActiveWindow();
	if (pWnd)
	{
		if (pWnd->IsKindOf(RUNTIME_CLASS(CWinGlkWndTextBuffer)))
			bEnable = TRUE;
	}
	pCmdUI->Enable(bEnable);
}

void CWinGlkMainWnd::OnFontProportional() 
{
	ChangeFont(false);
}

void CWinGlkMainWnd::OnFontFixed() 
{
	ChangeFont(true);
}

void CWinGlkMainWnd::OnOptions() 
{
	CGlkApp* pApp = (CGlkApp*)AfxGetApp();
	CPropertySheet OptionsDlg(IDS_OPTIONS,this);
	CWinGlkOptionGeneralPage GeneralPage;
	CWinGlkOptionStylePage StylePage;

	OptionsDlg.m_psh.dwFlags |= PSH_NOAPPLYNOW;
	OptionsDlg.AddPage(&GeneralPage);
	OptionsDlg.AddPage(&StylePage);

	GeneralPage.m_bBorders = pApp->GetWindowBorders();
	GeneralPage.m_bGUI = pApp->GetEnableGUI();
	GeneralPage.m_bStyleHints = pApp->GetStyleHints();
	GeneralPage.m_bSpeak = pApp->GetCanSpeak();
	GeneralPage.m_VoiceGUID = pApp->GetVoice();

	GeneralPage.m_bLinkColour = pApp->GetColourLinks();
	GeneralPage.m_bUnderlineColour = pApp->GetUnderlineLinks();
	GeneralPage.SetCustomLinkColour(pApp->GetCustomLinkColour());
	
	StylePage.SetCurrentStyle(wintype_TextBuffer,style_Normal);

	CWinGlkWnd* pActiveWnd = CWinGlkWnd::GetActiveWindow();
	if (OptionsDlg.DoModal() == IDOK)
	{
		pApp->SetSaveOptions(true);

		bool bBorderChanged =
			pApp->SetWindowBorders(GeneralPage.m_bBorders ? true : false);
		bool bGUIChanged =
			pApp->SetEnableGUI(GeneralPage.m_bGUI ? true : false);
		pApp->SetStyleHints(GeneralPage.m_bStyleHints ? true : false);

		pApp->SetCanSpeak(GeneralPage.m_bSpeak ? true : false);
		if (GeneralPage.m_VoiceGUID != pApp->GetVoice())
		{
			pApp->SetVoice(GeneralPage.m_VoiceGUID);
			TextToSpeech::GetSpeechEngine().Destroy();
		}

		pApp->SetColourLinks(GeneralPage.m_bLinkColour ? true : false);
		pApp->SetUnderlineLinks(GeneralPage.m_bUnderlineColour ? true : false);
		pApp->SetCustomLinkColour(GeneralPage.GetCustomLinkColour());

		if (bBorderChanged || bGUIChanged)
			GetView()->SizeWindows();

		// Redraw the display
		Invalidate();
	}
	if (pActiveWnd)
		pActiveWnd->SetActiveWindow();
}

void CWinGlkMainWnd::OnAbout() 
{
	CAboutDialog AboutDlg;
	AboutDlg.DoModal();
}

void CWinGlkMainWnd::StartTimer(int iMilliSecs)
{
	KillTimer(GlkTimer);

	if (iMilliSecs > 0)
		SetTimer(GlkTimer,iMilliSecs,NULL);
}

void CWinGlkMainWnd::SetBorders(bool bBorders)
{
	::SetWindowLong(m_View.GetSafeHwnd(),GWL_EXSTYLE,
		bBorders ? 0 : WS_EX_CLIENTEDGE);

	CWinGlkWnd* pWnd = NULL;
	do
	{
		pWnd = CWinGlkWnd::IterateWindows(pWnd,NULL);
		if (pWnd)
		{
			if (pWnd->GetSafeHwnd())
			{
				::SetWindowLong(pWnd->GetSafeHwnd(),GWL_EXSTYLE,
					bBorders ? WS_EX_CLIENTEDGE : 0);
			}
		}
	}
	while (pWnd);

	CRect WindowRect;
	GetWindowRect(WindowRect);
	MoveWindow(CRect(0,0,0,0),FALSE);
	MoveWindow(WindowRect,TRUE);
}

void CWinGlkMainWnd::SetGUI(bool bGUI)
{
	SetMenu(bGUI ? &m_Menus : NULL);
	if (m_UserToolBar.GetSafeHwnd())
		ShowControlBar(&m_UserToolBar,bGUI,TRUE);
	ShowControlBar(&m_GlkToolBar,bGUI,TRUE);
	ShowControlBar(&m_StatusBar,bGUI,TRUE);
	RecalcLayout(FALSE);
}

void CWinGlkMainWnd::EnableScrollback(bool bEnable)
{
	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu)
	{
		if (bEnable)
			pSysMenu->EnableMenuItem(IDM_SYS_SCROLLBACK,MF_BYCOMMAND|MF_ENABLED);
		else
			pSysMenu->EnableMenuItem(IDM_SYS_SCROLLBACK,MF_BYCOMMAND|MF_GRAYED);
	}
}

void CWinGlkMainWnd::ChangeFont(bool bFixed)
{
	CGlkApp* pApp = (CGlkApp*)AfxGetApp();
	CFontDialog FontDlg(
		bFixed ? pApp->GetFixedFont() : pApp->GetPropFont(),
		CF_SCREENFONTS|CF_NOSCRIPTSEL|(bFixed ? CF_FIXEDPITCHONLY : 0));
	if (FontDlg.DoModal() == IDOK)
	{
		GetView()->SizeWindows();
		Invalidate();
	}
	if (CWinGlkWnd::GetActiveWindow())
		CWinGlkWnd::GetActiveWindow()->SetActiveWindow();
}

void CWinGlkMainWnd::GetMessageString(UINT nID, CString& rMessage) const
{
	LPTSTR lpsz = rMessage.GetBuffer(255);
	if (AfxLoadString(nID,lpsz) != 0)
	{
		lpsz = _tcschr(lpsz,'\n');
		if (lpsz != NULL)
			*lpsz = '\0';
	}
	else
	{
		LPCTSTR lpszName = MAKEINTRESOURCE((nID>>4)+1);
		if (::FindResource(NULL,lpszName,RT_STRING) != NULL)
		{
			if (::LoadString(NULL,nID,lpsz,255) != 0)
			{
				lpsz = _tcschr(lpsz,'\n');
				if (lpsz != NULL)
					*lpsz = '\0';
			}
		}
	}
	rMessage.ReleaseBuffer();
}

CRgn CWinGlkMainWnd::m_Mask;

void CWinGlkMainWnd::SetWindowMask(CWinGlkGraphic* pGraphic)
{
	if (m_Mask.GetSafeHandle() == NULL)
	{
		CArray<RECT,RECT&> Regions;
		DWORD dwTransparent = 0x00FFFFFF;
		DWORD dwColour = 0;

		DWORD* ppvBits = (DWORD*)pGraphic->m_pPixels;
		int iHeight = abs(pGraphic->m_pHeader->biHeight);
		int iWidth = pGraphic->m_pHeader->biWidth;

		for (int y = 0; y < iHeight; y++)
		{
			int iLeft = -1;
			for (int x = 0; x < iWidth+1; x++)
			{
				// Get the pixel's colour
				if (x < iWidth)
					dwColour = CDibSection::GetPixel(ppvBits,iWidth,x,y);
				else
					dwColour = dwTransparent;

				if (iLeft >= 0)
				{
					if (dwColour == dwTransparent)
					{
						// Found the right hand edge of a rectangle in the region
						CRect Rect(iLeft,y,x,y+1);
						Regions.Add(Rect);
						iLeft = -1;
					}
				}
				else
				{
					if (dwColour != dwTransparent)
					{
						// Found the left hand edge of a rectangle in the region
						iLeft = x;
					}
				}
			}
		}

		if (Regions.GetSize() > 0)
		{
			// Create a RGNDATA structure
			int iDataSize = sizeof(RGNDATAHEADER)+(sizeof(RECT)*Regions.GetSize());
			RGNDATA* pRegionData = (RGNDATA*)new BYTE[iDataSize];
			::ZeroMemory(pRegionData,iDataSize);

			// Populate the RGNDATAHEADER
			pRegionData->rdh.dwSize = sizeof(RGNDATAHEADER);
			pRegionData->rdh.iType = RDH_RECTANGLES;
			pRegionData->rdh.nCount = Regions.GetSize();
			pRegionData->rdh.nRgnSize = 0;
			pRegionData->rdh.rcBound.right = iWidth;
			pRegionData->rdh.rcBound.bottom = iHeight;

			// Copy the RECTs into the RGNDATA
			::CopyMemory(((BYTE*)pRegionData)+sizeof(RGNDATAHEADER),
				Regions.GetData(),sizeof(RECT)*Regions.GetSize());

			// Create the region
			m_Mask.CreateFromData(NULL,iDataSize,pRegionData);
			delete[] pRegionData;
		}
	}
}
