/////////////////////////////////////////////////////////////////////////////
//
// Windows MFC GLK Libraries
//
// GlkSound
// GLK sound channels
//
/////////////////////////////////////////////////////////////////////////////

#ifndef WINGLK_SOUND_H_
#define WINGLK_SOUND_H_

extern "C"
{
#include "glk.h"
#include "gi_dispa.h"
}

class CWinGlkSound;

/////////////////////////////////////////////////////////////////////////////
// Class for Glk sound channels
/////////////////////////////////////////////////////////////////////////////

class CWinGlkSndChannel : public CObject
{
	DECLARE_DYNAMIC(CWinGlkSndChannel)

public:
	CWinGlkSndChannel(glui32 Rock);
	virtual ~CWinGlkSndChannel();

	glui32 GetRock(void);

	void SetDispRock(gidispatch_rock_t& Rock) { m_DispRock = Rock; };
	gidispatch_rock_t& GetDispRock(void) { return m_DispRock; };

	void SetArrayRock(gidispatch_rock_t& Rock) { m_ArrayRock = Rock; };
	gidispatch_rock_t& GetArrayRock(void) { return m_ArrayRock; };

	bool Play(CWinGlkSound* pSound, glui32 iSound, int iRepeat, int iNotify);
	void Stop(void);
	void SetVolume(int iVolume);
	void TimerPulse(void);

protected:
	glui32 m_Rock;
	gidispatch_rock_t m_DispRock;
	gidispatch_rock_t m_ArrayRock;

	CWinGlkSound* m_pSound;
	glui32 m_iSound;
	int m_iVolume;
	int m_iNotify;

/////////////////////////////////////////////////////////////////////////////
// Static data and member functions

public:
	static void CloseAllChannels(void);

	static bool IsValidChannel(CWinGlkSndChannel* pChannel);
	static CWinGlkSndChannel* IterateChannels(CWinGlkSndChannel* pChannel, glui32* pRockPtr);

protected:
	static CMap<CWinGlkSndChannel*,CWinGlkSndChannel*,int,int> ChannelMap;
};

/////////////////////////////////////////////////////////////////////////////
// Class for loaded sounds
/////////////////////////////////////////////////////////////////////////////

class CWinGlkSound
{
public:
	CWinGlkSound(BYTE* pData, int iLength);
	CWinGlkSound(LPCTSTR pszFileName);
	virtual ~CWinGlkSound();

	virtual bool Play(int iRepeat, int iVolume) = 0;
	virtual bool IsPlaying(void) = 0;
	virtual void SetVolume(int iVolume) = 0;

protected:
	double DecibelVolume(int iVolume);

protected:
	BYTE* m_pData;
	int m_iLength;
	bool m_bDelete;

	int m_iRepeat;
};

/////////////////////////////////////////////////////////////////////////////
// Base class for sound loaders
/////////////////////////////////////////////////////////////////////////////

class CWinGlkSoundLoader
{
public:
	CWinGlkSoundLoader() {};
	virtual ~CWinGlkSoundLoader() {};

public:
	// Get file prefix for sounds supported by this loader
	virtual LPCTSTR GetFilePrefix(void) = 0;

	// Get the file extensions for sounds supported for this loader
	virtual int GetNumberFileExtensions(void) = 0;
	virtual LPCTSTR GetFileExtension(int iExtIndex) = 0;

	// Get the identifier for sounds supported for this loader
	virtual glui32 GetIdentifier(void) = 0;
	
	// Get a sound object
	virtual CWinGlkSound* GetSound(LPCTSTR pszFileName) = 0;
	virtual CWinGlkSound* GetSound(BYTE* pData, int iLength) = 0;

	// Clean up after sound playback has been stopped
	virtual void SoundStopped(void) {};

/////////////////////////////////////////////////////////////////////////////
// Static data and member functions

public:
	// Set up the loaders
	static void InitLoaders(void);

	// Notify loaders to clean up after sound playback has been stopped
	static void AllSoundStopped(void);

	// Delete the loaders
	static void RemoveLoaders(void);

	// Get the number of loaders
	static int GetLoaderCount(void) { return m_Loaders.GetSize(); }

	// Get a loader by index
	static CWinGlkSoundLoader* GetLoader(int iIndex) { return m_Loaders.GetAt(iIndex); }

	// Get a loader for a given sound identifier
	static CWinGlkSoundLoader* GetLoaderForID(glui32 id);

protected:
	static CArray<CWinGlkSoundLoader*,CWinGlkSoundLoader*> m_Loaders;
};

#endif // WINGLK_SOUND_H_
