/////////////////////////////////////////////////////////////////////
// Simple speech interface for Glk
/////////////////////////////////////////////////////////////////////

#include "StdAfx.h"

#include <initguid.h>
#include "GlkTalk.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// The only instance of the speech object
TextToSpeech TextToSpeech::SpeechEngine;

// Get the only instance of the speech engine
TextToSpeech& TextToSpeech::GetSpeechEngine(void)
{
	return SpeechEngine;
}

// Constructor
TextToSpeech::TextToSpeech()
{
	m_iAvailable = -1;
	m_bInitialized = false;
	m_pCTTSMode = NULL;
}

// Destructor
TextToSpeech::~TextToSpeech()
{
	Destroy();
}

// Check if the engine is available
bool TextToSpeech::IsAvailable(void)
{
	if (m_iAvailable == -1)
	{
		CTTSEnum ModeEnum;
		m_iAvailable = SUCCEEDED(ModeEnum.Init()) ? 1 : 0;
	}
	return (m_iAvailable == 1);
}

// Enumerate the modes of the speech engine
const TextToSpeech::ModeInfoArray& TextToSpeech::GetModes(void)
{
	if (m_Modes.GetSize() == 0)
	{
		CTTSEnum ModeEnum;
		if (SUCCEEDED(ModeEnum.Init()))
		{
			TTSMODEINFOW ModeInfo;
			while (ModeEnum.Next(&ModeInfo) == S_OK)
			{
				struct ModeInfo Info;
				Info.Mode = ModeInfo.gModeID;
				Info.Name = ModeInfo.szModeName;
				m_Modes.Add(Info);
			}
		}
	}
	return m_Modes;
}

// Initialize the speech engine
void TextToSpeech::Initialize(GUID& Voice, LPUNKNOWN pIDirectSound)
{
	if (m_bInitialized)
		return;
	m_bInitialized = true;

	// Create an object to interface with the speech engine
	if (m_pCTTSMode == NULL)
	{
		m_pCTTSMode = new CTTSMode;
		if (m_pCTTSMode != NULL)
		{
			if (pIDirectSound != NULL)
				m_pCTTSMode->InitAudioDestDirect(pIDirectSound);

			HRESULT hr = m_pCTTSMode->Init(Voice);
			if (FAILED(hr))
				hr = m_pCTTSMode->Init();
			if (SUCCEEDED(hr))
			{
				// Get the mode GUID
				TTSMODEINFOW ModeInfo;
				if (SUCCEEDED(m_pCTTSMode->ModeGet(&ModeInfo)))
					Voice = ModeInfo.gModeID;
			}
		}
	}
}

// Close the speech engine
void TextToSpeech::Destroy(void)
{
	if (m_pCTTSMode != NULL)
	{
		delete m_pCTTSMode;
		m_pCTTSMode = NULL;
	}
	m_bInitialized = false;
}

// Speak some text
void TextToSpeech::Speak(LPCSTR speech)
{
	if (m_pCTTSMode != NULL)
	{
		int length = strlen(speech);

		// Convert the text to Unicode
		LPWSTR unicode = new WCHAR[length+1];
		::ZeroMemory(unicode,(length+1)*sizeof(WCHAR));

		if (::MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,speech,length,unicode,length+1) > 0)
			m_pCTTSMode->Speak(unicode,TRUE);

		// Free the Unicode buffer
		delete unicode;
	}
}
